﻿--[[ Copyright 2008 Erich Blume
    This file is part of Socialist.

    Socialist is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Socialist is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Socialist.  If not, see <http://www.gnu.org/licenses/>.
]]--
---------------------------------------
-- CONSTANTS                         --
---------------------------------------

-- Seconds between updates per char. If it hasn't been this long for this unit, no update will occur.
local MIN_UPDATE_INTERVAL = 300

-- Seconds between passive /who scans. If less than or equal the /who rate coded by WoW, will absolutely
-- cause thrashing (though WhoLib's API should handle thrashing gracefully).
-- TODO: Use that Queueing Theory I learned to annalyze this. Should be a D/M/1 queue.
local MIN_WHO_INTERVAL = 40

-- Level informaton. Unknown Level is a special value, -1
local UNKNOWN_LEVEL = -1
local MIN_LEVEL = UNKNOWN_LEVEL
local MAX_LEVEL = 255 -- GMs can be level 255.

---------------------------------------
-- Local Hooks                       --
---------------------------------------
-- This somewhat senseless section is for optimization. Ignore it.
local tremove = _G["tremove"]
local tinsert = _G["tinsert"]
local time = _G["time"]



---------------------------------------
-- Helper Functions & Lua Extensions --
---------------------------------------

local function tcount(tab)
	if type(tab) ~= "table" then return 0 end
	local i = 0
	for _ in pairs(tab) do
		i = i + 1
	end
	return i
end


---------------------------------------
-- Variable Definitions              --
---------------------------------------	
Socialist = LibStub("AceAddon-3.0"):NewAddon("Socialist",
	"AceConsole-3.0",
	"AceEvent-3.0"
	)
	
local wholib = LibStub:GetLibrary('LibWho-2.0'):Library()


local options = {
	name = "Socialist",
	handler = Socialist,
	type = "group",
	args = {
	
		cleardb = {
			name = "Clear Database",
			desc = "Clears Socialist's cache of people you've seen.",
			confirm = true,
			confirmText = "Are you sure you want to clear the database?",
			disabled = "IsDBEmpty",
			type = "execute",
			func = "ClearDB",
		},
		
		search = {
			name = "Search",
			desc = "Search Socialist's database for people you've seen.",
			type = "input",
			get = nil, -- "get" doesn't make sense here.
			set = function(info, value) Socialist:ConsoleSearch(value) end,
			multiline = false,
		},
	}
}

local defaults = {
	realm = {
		listing = {

		},	
	},
}

local _G = getfenv()

Socialist.MIN_LEVEL = MIN_LEVEL
Socialist.MAX_LEVEL = MAX_LEVEL

---------------------------------------
-- Events                            --
---------------------------------------

function Socialist:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("SocialistDB", defaults, "realm")
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Socialist",options,{"socialist"})
	self.timeFrame = CreateFrame("Frame",nil,nil)
end

function Socialist:OnEnable()
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT","OnMouseover")
	self:RegisterEvent("CHAT_MSG_CHANNEL","OnChatMsg")
	wholib:RegisterWhoLibEvent('WHOLIB_QUERY_RESULT', 'OnWho', self)
	self.timeFrame:SetScript("OnUpdate", self.TimeTick)
	if not self.db.char.greeted then
		self.db.char.greeted = true
		self:Print("Loaded! Type /socialist for more information.")
	end
	self.callbacks = self.callbacks or LibStub("CallbackHandler-1.0"):New(Socialist)
end

function Socialist:OnDisable()
	self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")
	self:UnregisterEvent("CHAT_MSG_CHANNEL")
	-- TODO: Unregister WhoLib event. See http://www.wowace.com/projects/wholib/tickets/1-unregister-event/
	self.timeFrame:SetScript("OnUpdate", nil)
end

function Socialist:OnWho(event, query, results, complete, name)
	-- note that "name" here isn't needed. We want results[i].Name
	for k,v in pairs(results) do
		self:UpdateEntry(v.Name,"who",{
			['race']=v.Race,
			['class']=v.Class,
			['guild']=v.Guild,
			['level']=v.Level,
			['location']=v.Zone,
		})
	end
end

function Socialist:OnMouseover()
	local name, realm = UnitName("mouseover")
	
	if name ~= "Unknown Entity" and UnitIsPlayer("mouseover") then
		local guildName, guildRankName, guildRankIndex = GetGuildInfo("mouseover")
		self:UpdateEntry(name,"mouseover",{
			['race'] = UnitRace("mouseover"),
			['class'] = UnitClass("mouseover"),
			['guild'] = guildName,
			['level'] = UnitLevel("mouseover"),
			['location'] = GetZoneText(),
		})
	end -- else, we don't have a valid unit
end

function Socialist:OnChatMsg(evnet, msg,id,lang,chan,target,flag,zoneid,channum)
	-- TODO: Detect cross-realm communication.
	-- In the likely event that this can't be done, disable detection during BGs/Arenas for fear of
	-- invalidating the database.
	-- For now we can just allow this small amount of invalidation.
	
	self:UpdateEntry(id,"chat",{
		['location']=chan,
	})
	wholib:Who(self:FormulateQuery(id),{['queue'] = wholib.WHOLIB_QUEUE_SCANNING})
end

---------------------------------------
-- Public API                        --
---------------------------------------

-- From CallbackHandler
-- TODO: Describe the DB_UPDATED events API

-- :FormulateQuery([name[,guild[,zone[,class[,race[,low[, high]]]]]]])
-- Formulates a /who-able query string based on the arguments.
-- Set unknown/unneeded arguments to nil (except if high is set, when low is needed)
-- No return
function Socialist:FormulateQuery(name,guild,zone,class,race,low,high)
	-- Stub (redifined later)
end

-- :IsDBEmpty()
-- Returns nontrue if the DB is populated.
function Socialist:IsDBEmpty()
	return self:DBSize() == 0
end

-- :DBSize()
-- Returns an int, minimum 0. See :IsDBEmpty()
function Socialist:DBSize()
	return tcount(self.db.realm.listing)
end

-- :GetEntry(name)
-- Returns a database entry (a table) of 'name', or nil otherwise.
-- Intended to be used in order to allow other addons/modules to modify the db,
--	so please feel free to change whatever values in this table you like.
-- Please use this INSTEAD of accessing Socialist.db.realm.listing directly,
-- incase the database format changes later.
function Socialist:GetEntry(name)
	-- Static return of one entry.
	-- Can return nil.
	-- Allows direct access to individual entries so that other addons may modift
	return Socialist.db.realm.listing[name]
end

-- :ClearEntry(entry)
-- Removes the specified entry (as returned from :GetEntry, or any one entity from
-- the table returned by :Query) from Socialist's database. If the entry does not
-- exist, the database will not change but DB_UPDATED will still be called (for
-- the conveniance of the module which called ClearEntry)
function Socialist:ClearEntry(entry)
	self.db.realm.listing[entry.name] = nil
	self.callbacks:Fire("DB_UPDATED")
end

-- :SignalEntryModified()
-- No return
-- Signals to Socialist that the database has been modified. Not a requirement
-- that it be called, but it is considered "polite" as it will cause Socialist
-- to trigger the DB_UPDATED event.
function Socialist:SignalEntryModified()
	self.callbacks:Fire("DB_UPDATED")
end

-- :Query(query[,database])
-- Returns a serialized result subset of the query upon the database.
--	query (table) -
--		[minLevel] -(see maxLevel)
--		[maxLevel] -If both not nil, then the returned entries' level must be
--					between these two values. Setting minLevel to an integer
--					will exclude entries of any level below minLevel. Similar
--					outcome for maxLevel but no minLevel. The defined range is
--					from Socialist.MIN_LEVEL to Socialist.MAX_LEVEL, though any
--					integer *should* work fine.
--		[guild] -	If a string, then the returned entries' guild must match
--					that string.
--		[name] -	If a string, then the returned entries' name must match that
--					string. See [normalize].
--		[normalize]-If nonfalse (e.g. false, nil), then the [name] query may
--					also match using the special "normalizeName" feature, which
--					will match common letters to their less-common UTF-8 extended
--					latin counterparts (which are a common pain in the ass on enUS
--					servers.) This feature is not guarunteed to work, and makes no
--					sense in many localizations (in which the normalize feature will
--					simply do nothing). In some cases, can as much as double the
--					search time, though usually it won't make much difference.
--  database (table) - The database to be searched. Safe to omit in 99% of cases.
function Socialist:Query(query,db)
	-- Stub (redefined later)
end

---------------------------------------
-- Methods and Functions             --
---------------------------------------

function Socialist:WhoScan()
	self.db.realm.listing = self.db.realm.listing or {}
	local size = tcount(self.db.realm.listing)
	if size > 0 then
		local rand = math.random(size)
		local itor = 1
		local entry = nil
		for k,v in pairs(self.db.realm.listing) do
			if itor == rand then entry = v end
			itor = itor + 1
		end
		assert(entry,"'Impossible' index error proved possible.")
		-- TODO: ensure the above error is never called. Delete it.
		wholib:Who(self:FormulateQuery(entry.name),{['queue'] = wholib.WHOLIB_QUEUE_SCANNING})
	end
end

function Socialist.TimeTick(frame,elapsed)
	local self = Socialist -- annoying but required.
	-- Who scan
	self.lastWhoScan = (self.lastWhoScan or 0) + elapsed
	if self.lastWhoScan >= MIN_WHO_INTERVAL then
		self.lastWhoScan = 0
		self:WhoScan()
	end
end

function Socialist:ConsoleSearch(search)
	-- MeatHoof: 77 Tauren Rogue
	local query = { name = search }
	for k,v in pairs(self:Query(query,self.db.realm.listing)) do
		self:Print(k ..": ".. v.level .. " " .. v.race .. " " .. v.class)
	end
end

function Socialist:AddZoneLocation(places,zone, moment)
	while #(places) >= 10 do tremove(places) end
	tinsert(places,1,{["date"]=moment,["place"]=zone})
end

function Socialist:ClearDB()
	self:Print("Clearing Database...")
	self.db.realm.listing = {}
	self.callbacks:Fire("DB_UPDATED")
end

function Socialist:UpdateEntry(name, source, info)
-- Name: Self explanatory. Need not be unique. Cross-realm covered in info.
-- Source: the unique data-source name, to avoid too-frequent updates.
-- Structure for info table (all fields optional):
-- {
--   ['realm'] = Player's realm. Use nil or "" if unimportant (IE same realm as player.)
--   ['race'] = Player's race.
--   ['class'] = Player's class.
--   ['level'] = Player's level, integer, use UNKNOWN_LEVEL for a KNOWN level of -1 (or ??), nil otherwise.
--   ['guild'] = Player's guild.
--   ['location'] = Location seen. Could be a zone, or a concept (IE "Trade Chat", "Guild Roster", etc.)
-- }
	if name == "" then return end
	local db = self.db.realm.listing
	local id = name
	if info and info.realm then id = name..'-'..info.realm end
	db[id] = db[id] or {}
	local entry = db[id]
	entry.lastseen = entry.lastseen or {}
	
	local now = time()
	
	local lastseen = entry.lastseen[source] or 0
	if lastseen + MIN_UPDATE_INTERVAL < now then
		-- Note that a simple export of info onto entry is not used as some fields
		-- are more complex than info takes into account for.
		entry['name'] = name
		entry['normalized'] = self:NormalizeName(name)
		entry['realm'] = info.realm
		entry['race'] = info.race
		entry['class'] = info.class
		entry['guild'] = info.guild
		entry['level'] = info.level
		if info.location then
			entry['places'] = entry['places'] or {}
			self:AddZoneLocation(entry['places'],info.location,now)
		end
		entry.lastseen[source] = now
		
		-- Callbacks
		self.callbacks:Fire("DB_UPDATED")
	end
end

function Socialist:NormalizeName(name)
	-- TODO: Allow this to be turned on/off. Default based on locale. (on for enUS, off else)
	return name:gsub(".",self:GetMatchTable(GetLocale())):upper()
end

function Socialist:GetMatchTable(locale)
-- TODO: Move this to an external localization file.
-- Note that capitilzation doesn't really matter as NormalizeName() is the expected caller, which
-- capitalizes every character anyway. Still, try and stick to proper capitalization in case of change.

	if (locale == "enUS") then
		-- TODO: Add to this list as I find creepyness in the database.
		return {
			['¢']	=	'c',
			['¥']	=	'Y',
			['©']	=	'c',
			['ª']	=	'a',
			['®']	=	'R',
			['µ']	=	'u',
			['À']	=	'A',
			['Á']	=	'A',
			['Â']	=	'A',
			['Ã']	=	'A',
			['Ä']	=	'A',
			['Å']	=	'A',
			['Æ']	=	'AE',
			['Ç']	=	'C',
			['È']	=	'E',
			['É']	=	'E',
			['Ê']	=	'E',
			['Ë']	=	'E',
			['Ì']	=	'I',
			['Í']	=	'I',
			['Î']	=	'I',
			['Ï']	=	'I',
			['Ð']	=	'D',
			['Ñ']	=	'N',
			['Ò']	=	'O',
			['Ó']	=	'O',
			['Ô']	=	'O',
			['Õ']	=	'O',
			['Ö']	=	'O',
			['×']	=	'x',
			['Ø']	=	'O',
			['Ù']	=	'U',
			['Ú']	=	'U',
			['Û']	=	'U',
			['Ü']	=	'U',
			['Ý']	=	'Y',
			['Þ']	=	'p',
			['ß']	=	'B',
			['à']	=	'a',
			['á']	=	'a',
			['â']	=	'a',
			['ã']	=	'a',
			['ä']	=	'a',
			['å']	=	'a',
			['æ']	=	'ae',
			['ç']	=	'c',
			['è']	=	'e',
			['é']	=	'e',
			['ê']	=	'e',
			['ë']	=	'e',
			['ì']	=	'i',
			['í']	=	'i',
			['î']	=	'i',
			['ï']	=	'i',
			['ð']	=	'o',
			['ñ']	=	'n',
			['ò']	=	'o',
			['ó']	=	'o',
			['ô']	=	'o',
			['õ']	=	'o',
			['ö']	=	'o',
			['ø']	=	'o',
			['ù']	=	'u',
			['ú']	=	'u',
			['û']	=	'u',
			['ü']	=	'u',
			['ý']	=	'y',
			['þ']	=	'b',
			['ÿ']	=	'y',
			-- TODO: The other 3 latin code pages.
		}
	else
		return {}
	end
end

function Socialist:Query(query,db)
	-- Recursive function, pruning db as it prunes query.
	-- Pruning method does quick searches first.
	-- Needs to be as fast as possible! Called often.
	-- TODO: Refactor, using the documented contract. Escape conditions and such
	--	have made this code insanely insane.
	if type(db) == "nil" then db = self.db.realm.listing end
	
	local result = {}
	if type(query.minLevel) ~= "nil"  or type(query.maxLevel) ~= "nil" then
		local minLevel = query.minLevel or MIN_LEVEL
		query.minLevel = nil -- prune
		local maxLevel = query.maxLevel or MAX_LEVEL
		query.maxLevel = nil -- prune
		if minLevel > maxLevel then return {} end -- logic escape condition (see set theory)
		if minLevel == MIN_LEVEL and maxLevel == MAX_LEVEL then return self:Query(query,db) end
		for k,v in pairs(db) do
			if v.level and v.level >= minLevel and v.level <= maxLevel then
				result[k] = v
			end
			-- Rejected Entities:
			--  * Unrecorded level (UNKNOWN_LEVEL is not necessarily rejected!)
			--  * Level out of range
		end
		return self:Query(query,result)
	elseif type(query.guild) ~= "nil" then
		local guild = query.guild -- TODO: Can this be ""?
		query.guild = nil -- prune
		local num = 0
		if not guild then return self:Query(query,db) end -- TODO: See above. Is this called ever?
		for k,v in pairs(db) do
			if v.guild and v.guild:find(guild,1,true) then
				result[k] = v
			end
		end
		return self:Query(query,result)
	elseif type(query.name) ~= "nil" then
		local name = query.name
		query.name = nil -- prune
		if not name then return self:Query(query,db) end
		for k,v in pairs(db) do
			if v.name:upper():find(name:upper(),1,true) or 
				(query.normalize and v.normalized:find(name:upper(),1,true)) 
				then
				-- TODO: Optimize this. Can we chuck out :find? Can we recurse to prune? How expensive is :find?
				result[k] = v
			end
		end
		
		return self:Query(query,result) -- Recurse
	else
		local result = {}
		local itor = 1
		for k,v in pairs(db) do
			result[itor] = v
			itor = itor + 1
		end
		return result
	end
end

function Socialist:FormulateQuery(name,guild,zone,class,race,low,high)
	if not name then name = "" else name = 'n-"' .. name ..'" ' end
	if not guild then guild = "" else guild = 'g-"' .. guild ..'" ' end
	if not zone then zone = "" else zone = 'z-"' .. zone ..'" ' end
	if not class then class = "" else class = 'c-"' .. class ..'" ' end
	if not race then race = "" else race = 'r-"' .. race .. '" ' end
	local level = ""
	if low and high then level = low..'-'..high elseif low then level = low end
	return name ..guild..zone..class..race..level
end




